/*
Hibernus++ for MSP430FR5739:

Hibernus++: Software-based approach to hibernate and restore adaptively the system's state
in response to a power failure. This software exploits an external comparator.

Citation: If you are using this code for your research please cite:

[1]	D. Balsamo, A. S. Weddell, G. V. Merrett, B. M. Al-Hashimi, D. Brunelli and L. Benini,
"Hibernus: Sustaining Computation During Intermittent Supply for Energy-Harvesting Systems,"
in IEEE Embedded Systems Letters, vol. 7, no. 1, pp. 15-18, March 2015.

[2]	D. Balsamo; A. Weddell; A. Das; A. Arreola; D. Brunelli; B. Al-Hashimi; G. Merrett; L. Benini,
"Hibernus++: A Self-Calibrating and Adaptive System for Transiently-Powered Embedded Devices,"
in IEEE Transactions on Computer-Aided Design of Integrated Circuits and Systems. 

This research has been supported by the University of Southampton and the University of Bologna.
Copyright 2018, Domenico Balsamo, All rights reserved.
06/02/2019
*/

#include <msp430.h>
#include <stdio.h>
#include <string.h>
#include "hibernation.h"


#pragma SET_DATA_SECTION(".fram_vars")

unsigned long int *FRAM_write_ptr = (unsigned long int *) FRAM_START; // define a pointer that points the Start Address

//This pointer is used to copy the RAM
unsigned long int *RAM_copy_ptr = (unsigned long int *) RAM_START; // define a pointer for memory copy
unsigned long int *FLAG_V_monitor = (unsigned long int *) FLAG; // define a pointer for memory copy
unsigned long int *FLAG_V_monitor_2 = (unsigned long int *) FLAG_2; // define a pointer for memory copy
unsigned long int *FRAM_read_ptr = (unsigned long int *) FRAM_TEST_START;  // define a pointer that points the Start Address
//unsigned long int *FLAG_V_monitor_3 = (unsigned long int *) FLAG_3; // define a pointer for memory copy



unsigned long int *FLAG_interrupt = (unsigned long int *) INT; // define a pointer for memory copy
unsigned long int *V_monitor = (unsigned long int *) VMON; // define a pointer for memory copy
unsigned long int *V_monitor_2 = (unsigned long int *) VMON_2; // define a pointer for memory copy
unsigned long int *V_monitor_3 = (unsigned long int *) VMON_3; // define a pointer for memory copy


unsigned long int *Counter = (unsigned long int *) COUNT; // define a pointer for memory copy
unsigned long int *Counter_2 = (unsigned long int *) COUNT_2; // define a pointer for memory copy
//unsigned long int *Counter_3 = (unsigned long int *) COUNT_3; // define a pointer for memory copy

unsigned long int *VH_Calibr = (unsigned long int *) CALIBR; // define a pointer for memory copy
unsigned long int *CC_Check = (unsigned long int *) CHECK; // define a pointer for memory copy



int t;

//This pointer is used to copy the general registers
unsigned int *Memory_copy_ptr; // define a pointer for memory copy

//These pointers and variable are used to set the PC
unsigned long int *FRAM_pc = (unsigned long int *) PC;

unsigned long int* current_SP;
unsigned long int PC_values;

unsigned long data; //4 Byte

extern int counter;

//This array is used to restore the state
unsigned int Registers[230];

unsigned int *Reg_copy_ptr;

unsigned int counter_check_reg;

unsigned int temp;

unsigned int counter_check;


// extern unsigned int CC_check;

unsigned int prova;

int pro;

unsigned int set;

int c;

long int deltaV;

#pragma SET_DATA_SECTION()

const unsigned int vol[256]= {
		1966, 1973, 1979, 1986, 1992, 1999, 2006, 2012, 2019, 2025, 2032, 2038, 2045, 2051, 2058, 2065, 2071, 2078, 2084, 2091,
		2097, 2104, 2110, 2117, 2124, 2130, 2137, 2143, 2150, 2156, 2163, 2169, 2176, 2183, 2189, 2196, 2202, 2209, 2215, 2222,
		2228, 2235, 2242, 2248, 2255, 2261, 2268, 2274, 2281, 2287, 2294, 2301, 2307, 2314, 2320, 2327, 2333, 2340, 2346, 2353,
		2360, 2366, 2373, 2379, 2386, 2392, 2399, 2405, 2412, 2419, 2425, 2432, 2438, 2445, 2451, 2458, 2464, 2471, 2478, 2484,
		2491, 2497, 2504, 2510, 2517, 2523, 2530, 2537, 2543, 2550, 2556, 2563, 2569, 2576, 2582, 2589, 2596, 2602, 2609, 2615,
		2622, 2628, 2635, 2641, 2648, 2655, 2661, 2668, 2674, 2681, 2687, 2694, 2700, 2707, 2714, 2720, 2727, 2733, 2740, 2746,
		2753, 2759, 2766, 2773, 2779, 2786, 2792, 2799, 2805, 2812, 2818, 2825, 2832, 2838, 2845, 2851, 2858, 2864, 2871, 2877,
		2884, 2891, 2897, 2904, 2910, 2917, 2923, 2930, 2936, 2943, 2950, 2956, 2963, 2969, 2976, 2982, 2989, 2995, 3002, 3009,
		3015, 3022, 3028, 3035, 3041, 3048, 3054, 3061, 3068, 3074, 3081, 3087, 3094, 3100, 3107, 3113, 3120, 3127, 3133, 3140,
		3146, 3153, 3159, 3166, 3172, 3179, 3186, 3192, 3199, 3205, 3212, 3218, 3225, 3231, 3238, 3245, 3251, 3258, 3264, 3271,
		3277, 3284, 3290, 3297, 3304, 3310, 3317, 3323, 3330, 3336, 3343, 3349, 3356, 3363, 3369, 3376, 3382, 3389, 3395, 3402,
		3408, 3415, 3422, 3428, 3435, 3441,	3448, 3454, 3461, 3467, 3474, 3481, 3487, 3494, 3500, 3507, 3513, 3520, 3526, 3533,
		3540, 3546, 3553, 3559, 3566, 3572, 3579, 3585, 3592, 3599, 3605, 3612, 3618, 3625, 3631, 3638 };

const unsigned int gen[230]= {
		  0x100, 0x102, 0x104, 0x120, 0x12a, 0x130, 0x140, 0x144, 0x146, 0x150, 0x152, 0x154, 0x156, 0x15c,
		  0x160, 0x162, 0x164, 0x166, 0x168, 0x16a, 0x16c, 0x180, 0x182, 0x186, 0x188, 0x18a, 0x18c, 0x18e,
		  0x198, 0x19a, 0x19c, 0x19e, 0x1b0, 0x200, 0x202, 0x204, 0x206, 0x20a, 0x20c, 0x20e, 0x216, 0x218,
		  0x21a, 0x21c, 0x21e, 0x220, 0x222, 0x224, 0x226, 0x22a, 0x22c, 0x22e, 0x236, 0x238, 0x23a, 0x23c,
		  0x23e, 0x320, 0x322, 0x324, 0x326, 0x32a, 0x32c, 0x336, 0x340, 0x342, 0x344, 0x346, 0x350, 0x352,
		  0x354, 0x356, 0x360, 0x36e, 0x380, 0x382, 0x384, 0x386, 0x390, 0x392, 0x394, 0x396, 0x3a0, 0x3ae,
		  0x3c0, 0x3c2, 0x3c4, 0x3c6, 0x3d0, 0x3d2, 0x3d4, 0x3d6, 0x3e0, 0x3ee, 0x400, 0x402, 0x404, 0x406,
		  0x410, 0x412, 0x414, 0x416, 0x420, 0x42e, 0x440, 0x442, 0x444, 0x446, 0x450, 0x452, 0x454, 0x456,
		  0x460, 0x46e, 0x4a0, 0x4a2, 0x4a8, 0x4aa, 0x4ac, 0x4ae, 0x4b0, 0x4b2, 0x4b4, 0x4b6, 0x4b8, 0x4ba,
		  0x4bc, 0x4be, 0x4c0, 0x4c2, 0x4c4, 0x4c6, 0x4c8, 0x4ca, 0x4cc, 0x4ce, 0x4d0, 0x4d2, 0x4d4, 0x4d6,
		  0x4d8, 0x4da, 0x4dc, 0x4de, 0x4e0, 0x4e2, 0x4e4, 0x4e6, 0x4e8, 0x4ea, 0x4ec, 0x500, 0x502, 0x504,
		  0x506, 0x508, 0x50e, 0x510, 0x512, 0x516, 0x51a, 0x520, 0x522, 0x526, 0x52a, 0x530, 0x532, 0x536,
		  0x53a, 0X5a0, 0x5a2, 0x5a4, 0x5a6, 0x5c0, 0x5c2, 0x5c6, 0x5c8, 0x5ca, 0x5cc, 0x5ce, 0x5d0, 0x5d2,
		  0x5da, 0x5dc, 0x5de, 0x5e0, 0x5e2, 0x5e6, 0x5e8, 0x5ea, 0x5ec, 0x5ee, 0x5f0, 0x5f2, 0x5fa, 0x5fc,
		  0x5fe, 0x640, 0x642, 0x646, 0x648, 0x64a, 0x64c, 0x64e, 0x654, 0x656, 0x658, 0x65a, 0x65c, 0x65e,
		  0x660, 0x66a, 0x66c, 0x66e, 0x700, 0x702, 0x704, 0x706, 0x708, 0x70a, 0x712, 0x71a, 0x71c, 0x71e,
		  0x8c0, 0x8c2, 0x8c4, 0x8c6, 0x8cc, 0x8ce};


int cal=0;
long int CalValue_1=0;
long int CalValue_2=0;
long int CalValue_3=0;
long int CalValue_4=0;
long int Vmax=0;
long int Vmon=2000;

unsigned long int Counter_3=0;


unsigned int h=0;


extern volatile unsigned long int ADCResult;

unsigned int i, k;
unsigned int ModeAddress =  FRAM_START;
unsigned int StartAddress =  RAM_START;

unsigned int TX_Buffer_V1[4] = {0,0,0,0};

void SystemInit(void){

  GPIO_configuration();

  Clock_configuration();

  // Calibration

  if((*VH_Calibr)!=1){

		*VH_Calibr=0;  // It is the first time that the algorithm is running so we have to consider the calibration.

  }


  if((*VH_Calibr)==0){

	  // P3.6 Calibration Debugger
	 	  P3DIR |= BIT6;
	 	  P3OUT |= BIT6; // LED  ON

	 	  deltaV=0;

		  Calibration();

		  P3OUT &= ~BIT6; // LED 4 OFF

	}


  // Test

  // Vmonitor is the function for dinamically setting the external comparator threshold
  // at the minimun value before executing the test.

  Vmonitor();

  // P2.0 is used for generating the interrupt from the external analog comparator circuit


  //this t flag is to understand if the P2.0 is already 1.If it is, we don't have to put the system in LPM4

//  t=1;

  __delay_cycles(500); // (?) we need this delay ?


  t=0;

  t = (P2IN & 0x0001);

  if(t==0){

	    *FLAG_interrupt=1;

	    P2IFG = 0;                         // P2 IFG flag cleared
	    P2IES &= ~(BIT0);                  // P2.0 Lo/Hi edge interrupt P2OUT &= ~(BIT0);
	    P2IE |= BIT0;                      // P2.0 interrupt enabled
	    P2IFG = 0;                         // P2 IFG flag cleared

	    // Port Configuration

	    //This Pin P1.4 is not working properly. don't use.

		P1OUT &= ~(BIT4);
		P1DIR |= BIT4;

		// LPM4
	    __bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
	    __no_operation();


  }

  Test();



  if(CalValue_2 < CalValue_1){


   unsigned int vmonitor_3;

//    if(1){

	   Counter_3=0;

//     vmonitor_3=3600;  //minimun value

	   vmonitor_3=VMIN;  //minimun value


   // set up external Comparator = Vmonitor

   Set_V(vmonitor_3); // Vmonitor by using the external circuit

   __delay_cycles(500);

   t=1;

   t = (P2IN & 0x0001);

   if(t==0){  //the voltage is increasing slowly

		*FLAG_interrupt=3;


		P2IFG = 0;                         // P2 IFG flag cleared
		P2IES &= ~(BIT0);                  // P2.0 Lo/Hi edge interrupt P2OUT &= ~(BIT0);
		P2IE |= BIT0;                      // P2.0 interrupt enabled
		P2IFG = 0;                         // P2 IFG flag cleared

		// Port Configuration

		//This Pin P1.4 is not working properly. don't use.

		P1OUT &= ~(BIT4);
		P1DIR |= BIT4;

		StartDebounceTimer_2();

		// LPM4
		__bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
		__no_operation();


      }else{
    	  //the voltage is increasing quickly

    	  vmonitor_3=VMAX;  //minimun value

    	     // set up external Comparator = Vmonitor

    	   Set_V(vmonitor_3); // Vmonitor by using the external circuit

    	   __delay_cycles(500);

    	   t=1;

    	   t = (P2IN & 0x0001);

    	   if(t==0){  //the voltage is increasing slowly

    	  		*FLAG_interrupt=1;


    	  		P2IFG = 0;                         // P2 IFG flag cleared
    	  		P2IES &= ~(BIT0);                  // P2.0 Lo/Hi edge interrupt P2OUT &= ~(BIT0);
    	  		P2IE |= BIT0;                      // P2.0 interrupt enabled
    	  		P2IFG = 0;                         // P2 IFG flag cleared

    	  		// Port Configuration

    	  		//This Pin P1.4 is not working properly. don't use.

    	  		P1OUT &= ~(BIT4);
    	  		P1DIR |= BIT4;

    	  		StartDebounceTimer_2();

    	  		// LPM4
    	  		__bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
    	  		__no_operation();

    	   }

      }



  }


  if ((*CC_Check!=0) && (*CC_Check!=1) ){

//	  unsigned long int value= 3400 + deltaV;

	  *CC_Check=0;


	   unsigned int try = (unsigned int) (VMINN + deltaV);

//      unsigned long int value = VMIN2;


		Set_V(try);

		  __delay_cycles(500); // (?) we need this delay ?

		  *FLAG_interrupt=2;

		  P2IFG = 0;                         // P2 IFG flag cleared

		  P2IES |= BIT0;                     // P2.0 Hi/lo edge interrupt

		  P2IE |= BIT0;                      // P2.0 interrupt enabled
		  P2IFG = 0;                         // P2 IFG flag cleared

		   __bis_SR_register(GIE);        // LPM0, ADC10_ISR will force exit
		   __no_operation();


	  }



   if (*CC_Check==0){

			deltaV=deltaV+10;

			*CC_Check=0;

//			unsigned long int value_2 = 3400 + deltaV;

			 unsigned int value_2 =(unsigned int) (VMINN + deltaV);

			 *FLAG_interrupt=2;

//			unsigned long int value_2 = VMIN2;

			Set_V(value_2);

			__delay_cycles(500); // (?) we need this delay ?s

			P2IFG = 0;                         // P2 IFG flag cleared

			P2IES |= BIT0;                     // P2.0 Hi/lo edge interrupt

			P2IE |= BIT0;                      // P2.0 interrupt enabled
			P2IFG = 0;                         // P2 IFG flag cleared

			__bis_SR_register(GIE);        // LPM0, ADC10_ISR will force exit
			__no_operation();

	  }




   if (*CC_Check==1){

	   *CC_Check=0;

//	   unsigned long int value_3 = 3400 + deltaV;

	   unsigned long int value_3 =(unsigned int) (VMINN + deltaV);

//	   unsigned long int value_3 = VMIN2;

		Set_V(value_3);

	    P3DIR |= BIT0;
	    P3OUT |= BIT0; // LED  ON

		Restore();

	  }

   }


void Test(void){


  	  //ADC initialization for the test

  	  // Configure ADC10 - Pulse sample mode; ADC10SC trigger
  	  ADC10CTL0 = ADC10SHT_11 + ADC10ON;         // 16 ADC10CLKs; ADC ON

  	  ADC10CTL1 = ADC10SHP + ADC10CONSEQ_0;     // s/w trig, single ch/conv
  	  ADC10CTL2 = ADC10RES;                     // 10-bit conversion results
 //	  ADC10MCTL0 = ADC10SREF_1 + ADC10INCH_11;  // AVcc/2 internal divider for Vcc

  	  ADC10MCTL0 = ADC10SREF_3 + ADC10INCH_2;  // Veref+ and external divider - In A2

  	  // Configure internal reference
  	  while(REFCTL0 & REFGENBUSY);              // If ref generator busy, WAIT
  	  REFCTL0 |= REFVSEL_1+REFON;               // Select internal ref = 1.5V
  	                                            // Internal Reference ON
  	  __delay_cycles(400);                      // Delay for Ref to settle


	  ADC10CTL0 |= ADC10ENC + ADC10SC;        // Sampling and conversion start
	  while (ADC10CTL1 & ADC10BUSY);          // ADC10BUSY?
	  __no_operation();                       // For debug only

	  CalValue_1 = ADC10MEM0;


 //	  ADC10CTL0 &= ~ ADC10ON;        // 16 ADC10CLKs; ADC ON

     P1DIR |= BIT4;
     P1OUT |= BIT4;

	 test();

     P1OUT &= ~ BIT4;


	 CalValue_2=0;

	  ADC10CTL0 |= ADC10ENC + ADC10SC;        // Sampling and conversion start
	  while (ADC10CTL1 & ADC10BUSY);          // ADC10BUSY?
	  __no_operation();                       // For debug only

	 CalValue_2 = ADC10MEM0;




  	// ADC is turned off

  	  ADC10CTL0 &= ~ ADC10ON;        // 16 ADC10CLKs; ADC ON

  	 // only for debugging

  	  CalValue_1=(((CalValue_1*1179)/1024)*4200)/1000; //4152

  	  CalValue_2=(((CalValue_2*1179)/1024)*4200)/1000; //4152

        //Test_completed

  	  *FLAG_V_monitor=1;

  	  P1OUT &= ~ BIT7; // LED  ON


}




void GPIO_configuration (void){

    P2REN &= ~(BIT0);                  // Disable pull-down and pull-up resistor
    P2DIR &= ~(BIT0);                  // Direction = BIT0 input

	//GPIO configuration

	P1OUT &= ~(0xF9);
	P2OUT &= ~(0xFE);
	P3OUT &= ~(0xFB);
	P4OUT &= ~(0xFF);
	PJOUT &= ~(0xFF); // Configure the pull-down resistor


	P1DIR |= 0xF9;
	P2DIR |= 0xFE;
	P3DIR |= 0xFB;
	P4DIR |= 0xFF;
	PJDIR |= 0xFF; // // Direction = output


	//External reference (Vcc)

	P1SEL1 |= BIT1;
	P1SEL0 |= BIT1;

	//External ADC input

	P1SEL1 |= BIT2;
	P1SEL0 |= BIT2;

//	__delay_cycles(5000); // (?) we need this delay ?

	//This GPIO is used for the short-circuit

	P1OUT &= ~(BIT3);
	P1DIR |= BIT3;


}

void Clock_configuration (void){

	  // Startup clock system in max. DCO setting ~8MHz
	  // The first thing is to set all clocks of the system MCLK, ACLK, SMCLK by using the CS control registers.

	  // Note: The registers have 16bits. For a generic register, the suffix "_L" (ANYREG_L) refers to the lower byte
	  // of the register (bits 0 through 7). The suffix "_H" refers to the upper byte of the register(bits 8 through 15).

	  // The registers are locked, then, is need to use a pwd to unlock them.
	  // CSCTL0 Clock System Control Register 0 is only used to unlock the other CS registers used for clock setting.

	  // Allow for settling delay
	  //  __delay_cycles(10000);

	  //core frequency

	  CSCTL0_H = 0xA5;                          // Unlock register
	                                            //CSKEY password. Always reads as 0A5h. By writing any other value in
	                                            // word mode generates a PUC. After a correct password, CS register accesses are enabled.

	  //  CSCTL1 |= DCORSEL;  //Set max. DCO setting. 20MH in this case

	  CSCTL1 &= ~(DCORSEL);  //Set max. DCO setting. 5.33MH in this case

	  CSCTL1 |= DCOFSEL0 + DCOFSEL1;            //DCO frequency select register
	                                              //Set max. DCO setting. 8MH in this case

	//  CSCTL1 &= ~(DCOFSEL1);  //Set max. DCO setting. 5.33MH in this case
	//  CSCTL1 &= ~(DCOFSEL0);  //Set max. DCO setting. 5.33MH in this case

	  CSCTL2 = SELA_1 + SELS_3 + SELM_3;        // Selects the ACLK, MCLK AND SMCLK sources register
	                                            // set ACLK = VLO; MCLK = DCO; SMCLK = DCO;

	  CSCTL3 = DIVA_0 + DIVS_0 + DIVM_0;        // MCLK, SMCLK and ACLK source divider/prescaler register.

	  CSCTL0_H = 0x01;                          // Lock Register


}

void StartDebounceTimer(void)

 // Timer_A is a 16-bit timer/counter with up to seven capture/compare registers. Timer_A can support
 // multiple capture/compares, PWM outputs, and interval timing. Timer_A also has extensive interrupt
 // capabilities.

{

  // default delay = 0
  // Debounce time = 1500* 1/8000 = ~200ms
    TA1CCTL0 = CCIE;                          // TACCR0 interrupt enabled

//  TA1CCR0 = 750; // 88ms
    TA1CCR0 = 375; // 88ms

//  TA1CCR0 = 16000; // 88ms

  TA1CTL = TASSEL_1 + MC_1;                 // Timer_A clock source select TA1CTL registers: 01b = ACLK

}

void StartDebounceTimer_2(void)

 // Timer_A is a 16-bit timer/counter with up to seven capture/compare registers. Timer_A can support
 // multiple capture/compares, PWM outputs, and interval timing. Timer_A also has extensive interrupt
 // capabilities.

{

  // default delay = 0
  // Example: Debounce time = 1500* 1/8000 = 187.5ms

  TA1CCTL0 = CCIE;                          // TACCR0 interrupt enabled

//  TA1CCR0 = 800;  // Example: Debounce time = 800 * 1/8000 = 100ms
    TA1CCR0 = 80;  // Example: Debounce time = 80 * 1/8000 = 10ms


  TA1CTL = TASSEL_1 + MC_1;                 // Timer_A clock source select TA1CTL registers: 01b = ACLK

}



void test(void){

    while(FRAM_read_ptr < (unsigned long int *) (FRAM_TEST_END)) {

    	data=*FRAM_read_ptr++;

    }


}

/*

void FRAM_Read (unsigned int StartAddress)
{
  unsigned long *FRAM_read_ptr;  // define a pointer that points the Start Address
  unsigned int i=0;

  //In case FRAM Write Address is corrupted, exit loop
  if((StartAddress >= FRAM_TEST_START) && (StartAddress < FRAM_TEST_END))
  {
  	// Setup FRAM pointer
  	*FRAM_read_ptr = (unsigned long *) StartAddress;

  	// Write 128 * 4 = 512 bytes  //64 109us
  	for ( i= 0; i<640; i++)
  	  {
  		data=*FRAM_read_ptr++;
      }
  }
}

*/

void Vmonitor(void){



	unsigned int vmonitor;

	// If it is the first time *FLAG_V_monitor==FFFF

	if ((*FLAG_V_monitor!=0) && (*FLAG_V_monitor!=1) ){

		*V_monitor=VMIN; // Minimun value

//		*V_monitor=3400; //for testing

		*Counter=0; // Counter is for counting how many times the test fails
		            // Because the energy available is not enough to complete the test


	}



	// If the test fails

	if (*FLAG_V_monitor==0){

		*Counter=*Counter+1;

		*V_monitor=VMIN +((*Counter)*100);

//		*V_monitor=3400+((*Counter)*100);  //for testing


	}

	// If the test is good, *V_monitor conteins the right value for the next times

	vmonitor= (unsigned int) (*V_monitor);

	// STATE = 0

	*FLAG_V_monitor=0; //Flag is set to 0 when the test is starting.

    // P1.7 Test routine debugger pin
	P1DIR |= BIT7;
	P1OUT |= BIT7; // LED  ON

	  // set up external Comparator = vmonitor

    Set_V(vmonitor); // Vmonitor by using the external circuit


}


void Vmonitor_2(void){


	unsigned int vmon;


	if ((*FLAG_V_monitor_2!=0) && (*FLAG_V_monitor_2!=1) ){

		*V_monitor_2 = VMIN; //min. value

//		*V_monitor_2 = 3400; //for testing

		*Counter_2=0; // Counter is for counting how many times the test fails
		            // Because the energy available is not enough to complete the test

	}


	if (*FLAG_V_monitor_2==0){

		*Counter_2=*Counter_2+1;

		*V_monitor_2=VMIN +((*Counter)*100);

//		*V_monitor_2=3400+((*Counter_2)*100);  //for testing


	}


	vmon = (unsigned int) (*V_monitor_2);

	// STATE = 0

	*FLAG_V_monitor_2=0; //Flag is set to 0 when the test is starting.

	  // set up external Comparator = Vmonitor

    Set_V(vmon); // Vmonitor by using the external circuit


}



//}


void Calibration(void){


	  Vmonitor_2();

	  // P2.0 is used for generating the interrupt from the external analog comparator circuit

	  //this flag is to understand if the P2.0 is already 1.If it is, we don't have to put the system in LPM4

	  __delay_cycles(500);

	  t=1;

	  t = (P2IN & 0x0001);

	  if(t==0){

		    *FLAG_interrupt=1;

		    P2IFG = 0;                         // P2 IFG flag cleared

		    P2IES &= ~(BIT0);                  // P2.0 Lo/Hi edge interrupt P2OUT &= ~(BIT0);

		    P2IE |= BIT0;                      // P2.0 interrupt enabled

		    P2IFG = 0;                         // P2 IFG flag cleared

		    // Port Configuration

		    //This Pin P1.4 is not working properly. don't use.

			P1OUT &= ~(BIT4);
			P1DIR |= BIT4;

			// LPM4
		    __bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
		    __no_operation();


	  }


  	  //ADC initialization for the test

  	  // Configure ADC10 - Pulse sample mode; ADC10SC trigger
  	  ADC10CTL0 = ADC10SHT_11 + ADC10ON;         // 16 ADC10CLKs; ADC ON

  	  ADC10CTL1 = ADC10SHP + ADC10CONSEQ_0;     // s/w trig, single ch/conv
  	  ADC10CTL2 = ADC10RES;                     // 10-bit conversion results
 //	  ADC10MCTL0 = ADC10SREF_1 + ADC10INCH_11;  // AVcc/2 internal divider for Vcc

  	  ADC10MCTL0 = ADC10SREF_3 + ADC10INCH_2;  // Veref+ and external divider - In A2

  	  // Configure internal reference
  	  while(REFCTL0 & REFGENBUSY);              // If ref generator busy, WAIT
  	  REFCTL0 |= REFVSEL_1+REFON;               // Select internal ref = 1.5V
  	                                            // Internal Reference ON
  	  __delay_cycles(400);                      // Delay for Ref to settle


//  	  for(i=0; i<10; i++){


  	 ADC10CTL0 |= ADC10ENC + ADC10SC;        // Sampling and conversion start
  	 while (ADC10CTL1 & ADC10BUSY);          // ADC10BUSY?
  	 __no_operation();                       // For debug only

  	 CalValue_3 = ADC10MEM0;


  	 // GPIO to short-circuit the input

     P1DIR |= BIT3;
	 P1OUT |= BIT3;


	  ADC10CTL0 &= ~(ADC10ON);

	  CheckPoint_calibrate();

//      ADC10CTL0 = ADC10ON;


	  // Configure ADC10 - Pulse sample mode; ADC10SC trigger
	  ADC10CTL0 = ADC10SHT_11 + ADC10ON;         // 16 ADC10CLKs; ADC ON

	  ADC10CTL1 = ADC10SHP + ADC10CONSEQ_0;     // s/w trig, single ch/conv
	  ADC10CTL2 = ADC10RES;                     // 10-bit conversion results
//	  ADC10MCTL0 = ADC10SREF_1 + ADC10INCH_11;  // AVcc/2 internal divider for Vcc

	  ADC10MCTL0 = ADC10SREF_3 + ADC10INCH_2;  // Veref+ and external divider - In A2

	  // Configure internal reference
	  while(REFCTL0 & REFGENBUSY);              // If ref generator busy, WAIT
	  REFCTL0 |= REFVSEL_1+REFON;               // Select internal ref = 1.5V
	                                            // Internal Reference ON
	  __delay_cycles(400);                      // Delay for Ref to settle


	  CalValue_4 = 0;

//	  for(i=0; i<10;i++){

	  ADC10CTL0 |= ADC10ENC + ADC10SC;        // Sampling and conversion start
	  while (ADC10CTL1 & ADC10BUSY);          // ADC10BUSY?
	  __no_operation();                       // For debug only

	  CalValue_4 = ADC10MEM0;

//		  CalValue_3 = ADC10MEM0;

//	  }

	  P1OUT &= ~(BIT3);

//	  CalValue_3 = CalValue_3/10;

	  ADC10CTL0 &= ~(ADC10ON);


	  // only for debugging

	  CalValue_4=(((CalValue_4*1179)/1024)*4200)/1000; //4152

	  CalValue_3=(((CalValue_3*1179)/1024)*4200)/1000; //3  instead of 3.2 because of the op.amp.

	  deltaV= (long int) (CalValue_3 - CalValue_4);



	  *FLAG_V_monitor_2=1;
	  *VH_Calibr=1;


}



void Set_V (unsigned int v) {

    counter_check=0;

  	for (i= 0; i<256; i++)
  	  {
  		if(v>vol[i]){
  			counter_check++;
  		} else {
  			break;
  		}

   	}


    P4OUT &= ~(BIT0); //blue
    P2OUT &= ~(BIT1); //yellow
    P2OUT &= ~(BIT2); // green
  	P3OUT &= ~(BIT3);  //orange
    P1OUT &= ~(BIT0); //blue
    P1OUT &= ~(BIT5); //yellow
    P1OUT &= ~(BIT6); //green
    P2OUT &= ~(BIT5); //orange




//  	P4DIR |= BIT0;
    P4OUT |= (BIT0 & (counter_check & 0x0001)); //blue

//    P2DIR |= BIT1;
    P2OUT |= (BIT1 & (counter_check & 0x0002)); //yellow

//    P2DIR |= BIT2;
    P2OUT |= (BIT2 & (counter_check & 0x0004)); // green

//    P3DIR |= BIT3;
  	P3OUT |= (BIT3 & (counter_check & 0x0008));  //orange

//  	P1DIR |= BIT0;
    P1OUT |= (BIT0 & ((counter_check & 0x0010) >> 4)); //blue

//    P1DIR |= BIT5;
    P1OUT |= (BIT5 & (counter_check & 0x0020)); //yellow

//    P1DIR |= BIT6;
    P1OUT |= (BIT6 & (counter_check & 0x0040)); //green

//   P2DIR |= BIT5;
    P2OUT |= (BIT5 & ((counter_check & 0x0080) >> 2)); //orange


}


void CheckPoint (void) {

	P3DIR |= BIT5;
	P3OUT |= BIT5; // LED 4 ON

	P2OUT &= ~BIT6; // LED  OFF

	*CC_Check=0;


	// copy all the registers onto the FRAM

	          asm(" MOVA R1,&0xEDF4");
			  asm(" MOVA R2,&0xEDF8");
			  asm(" MOVA R3,&0xEDFC");
			  asm(" MOVA R4,&0xEE00");
			  asm(" MOVA R5,&0xEE04");
			  asm(" MOVA R6,&0xEE08");
			  asm(" MOVA R7,&0xEE0C");
			  asm(" MOVA R8,&0xEE10");
			  asm(" MOVA R9,&0xEE14");
			  asm(" MOVA R10,&0xEE18");
			  asm(" MOVA R11,&0xEE1C");
			  asm(" MOVA R12,&0xEE20");
			  asm(" MOVA R13,&0xEE24");
			  asm(" MOVA R14,&0xEE28");
			  asm(" MOVA R15,&0xEE2C");


			  current_SP = (void*) _get_SP_register();

			  *FRAM_pc= *current_SP;



     // copy all the RAM onto the FRAM

			  Save_RAM();

	 // copy all the general registers onto the FRAM

			  pro=0;


//			    P1DIR |= BIT0;                            // P1.0 output
//				P1OUT |= BIT0;

			  Save_Register();

//	    		P1OUT &= ~BIT0;



	    P3OUT &= ~BIT5; // LED 4 OFF
//	    PJOUT ^= BIT0; // LED  OFF

		*CC_Check = 1;



}

void CheckPoint_calibrate (void) {

//	P2DIR |= BIT5;
//	P2OUT |= BIT5; // LED 4 ON

//	CC_check=1;


	// copy all the registers onto the FRAM



	          asm(" MOVA R1,&0xEDF4");
			  asm(" MOVA R2,&0xEDF8");
			  asm(" MOVA R3,&0xEDFC");
			  asm(" MOVA R4,&0xEE00");
			  asm(" MOVA R5,&0xEE04");
			  asm(" MOVA R6,&0xEE08");
			  asm(" MOVA R7,&0xEE0C");
			  asm(" MOVA R8,&0xEE10");
			  asm(" MOVA R9,&0xEE14");
			  asm(" MOVA R10,&0xEE18");
			  asm(" MOVA R11,&0xEE1C");
			  asm(" MOVA R12,&0xEE20");
			  asm(" MOVA R13,&0xEE24");
			  asm(" MOVA R14,&0xEE28");
			  asm(" MOVA R15,&0xEE2C");


			  current_SP = (void*) _get_SP_register();

			  *FRAM_pc= *current_SP;



     // copy all the RAM onto the FRAM

			  Save_RAM();

	 // copy all the general registers onto the FRAM

//			  pro=0;


//			    P1DIR |= BIT0;                            // P1.0 output
//				P1OUT |= BIT0;

			  Save_Register();

//	    		P1OUT &= ~BIT0;



//	    P2OUT &= ~BIT5; // LED 4 OFF
//	    PJOUT ^= BIT0; // LED  OFF

//		CC_check = 0;


}


void Restore (void){

    	// Restore all Register and RAM

//    P4DIR |= BIT0;
//    P4OUT |= BIT0; // LED  ON

//    P2DIR |= BIT6;
//	  P2OUT |= BIT6; // LED  ON

	             Restore_Register();

	             // Restore RAM
	              FRAM_write_ptr= (unsigned long int *) FRAM_START;

	              	for ( i= 0; i<16; i++)
	              	  {
	              		*FRAM_write_ptr++;
	              			      }

	                RAM_copy_ptr= (unsigned long int *) RAM_START;

	                while(RAM_copy_ptr < (unsigned long int *) (RAM_END)) {

	                		*RAM_copy_ptr++=*FRAM_write_ptr++;

	                }

	              asm(" MOVA &0xEDF4,R1");
	              asm(" MOVA &0xEDF8,R2");
    			  asm(" MOVA &0xEDFC,R3");
    			  asm(" MOVA &0xEE00,R4");
    			  asm(" MOVA &0xEE04,R5");
    			  asm(" MOVA &0xEE08,R6");
    			  asm(" MOVA &0xEE0C,R7");
    			  asm(" MOVA &0xEE10,R8");
    			  asm(" MOVA &0xEE14,R9");
    			  asm(" MOVA &0xEE18,R10");
    			  asm(" MOVA &0xEE1C,R11");
    			  asm(" MOVA &0xEE20,R12");
    			  asm(" MOVA &0xEE24,R13");
    			  asm(" MOVA &0xEE28,R14");
    			  asm(" MOVA &0xEE2C,R15");

    			  *current_SP=*FRAM_pc;

    			  P3OUT &= ~BIT0; // LED  ON


    			  __delay_cycles(500); // (?) we need this delay ?

    			  *FLAG_interrupt=2;

    			  P2IFG = 0;                         // P2 IFG flag cleared

    			  P2IES |= BIT0;                     // P2.0 Hi/lo edge interrupt

    			  P2IE |= BIT0;                      // P2.0 interrupt enabled
    			  P2IFG = 0;                         // P2 IFG flag cleared

    			   __bis_SR_register(GIE);        // LPM0, ADC10_ISR will force exit
    			   __no_operation();

    			  *CC_Check=0;

    			  pro=1;

    		      P2DIR |= BIT6;
    		      P2OUT |= BIT6; // LED  ON

    }


void Save_RAM (void){

	FRAM_write_ptr= (unsigned long int *) FRAM_START;

	for ( i= 0; i<16; i++)
	{
		*FRAM_write_ptr++;
		      }


// copy all RAM onto the FRAM

  RAM_copy_ptr= (unsigned long int *) RAM_START;


	  while(RAM_copy_ptr <= (unsigned long int *) (RAM_END)) {
		  *FRAM_write_ptr++ =*RAM_copy_ptr++;
	  }


	}

/*

extern void Restore_Comparator_State(void){


	P1DIR |= 0xff; P2DIR |= 0xff; P3DIR |= 0xff; P4DIR |= 0xff;PJDIR |= 0xff;
	P1OUT &= ~(0xff); P2OUT &= ~(0xff);P3OUT &= ~(0xff);P4OUT &= ~(0xff);PJOUT &= ~(0xff);

	P3DIR |= BIT7;
	P3OUT |= BIT7; // LED  ON

	P3DIR |= BIT4;
	P3OUT |= BIT4; // LED  ON

    P2OUT |= BIT0;                     // Configure pull-up resistor
    P2DIR &= ~(BIT0);                  // Direction = BIT0 input
    P2REN |= BIT0;                     // Enable pull-up resistor
    P2IES &= ~(BIT0);                     // P2.0 Lo/Hi edge interrupt
    P2IE |= BIT0;                      // P2.0 interrupt enabled
    P2IFG = 0;                         // P2 IFG flag cleared


    Set_V(3000);

	P3OUT &= ~BIT4; // LED  ON

	__bis_SR_register(LPM4_bits+GIE); // Enter LPM4 with inetrrupts enabled
	__no_operation();             // For debug

}


*/


void Save_Register (void){

		for(i=0;i<230;i++){

				Reg_copy_ptr=(unsigned  int	*)	gen[i];
				Registers[i]=*Reg_copy_ptr;
	    }


	}


void Restore_Register (void){

	MPUCTL0_H = 0xA5;
	PMMCTL0_H =  0xA5;
	FRCTL0_H = 0xA5;
	CSCTL0_H = 0xA5;

	for(i=0;i<3;i++){

			Reg_copy_ptr=(unsigned  int	*)	gen[i];
			*Reg_copy_ptr=Registers[i];

	}


	for(i=4;i<6;i++){

			Reg_copy_ptr=(unsigned  int	*)	gen[i];
			*Reg_copy_ptr=Registers[i];

	}

	for(i=7;i<13;i++){

			Reg_copy_ptr=(unsigned  int	*)	gen[i];
			*Reg_copy_ptr=Registers[i];

	}


	for(i=15;i<34;i++){

		    Reg_copy_ptr=(unsigned  int	*)	gen[i];
			*Reg_copy_ptr=Registers[i];

	}

	for(i=36;i<46;i++){

		    Reg_copy_ptr=(unsigned  int	*)	gen[i];
			*Reg_copy_ptr=Registers[i];

	}

	for(i=48;i<169;i++){

		    Reg_copy_ptr=(unsigned  int	*)	gen[i];
			*Reg_copy_ptr=Registers[i];

	}


	for(i=170;i<230;i++){

		    Reg_copy_ptr=(unsigned  int	*)	gen[i];
			*Reg_copy_ptr=Registers[i];

	}

	MPUCTL0_H = 0x01;
	PMMCTL0_H = 0x01;
	FRCTL0_H = 0x01;
	CSCTL0_H = 0x01;

}



#pragma vector=PORT2_VECTOR
__interrupt void Port_2(void)
{

  switch(__even_in_range(P2IV,P2IV_P2IFG0)) // I think that even_in_range is correlated to the INT considered in a given Port
    {
     case P2IV_P2IFG0:

    	 //disable the interrupt for the timer

         TA1CCTL0 = 0;  // Disable Interrupt
    	 TA1CTL = 0;    // Disable Timer

		 // Toggles interrupt edge

//		 P2IES ^= BIT0;

		 P2IFG = 0;
		 P2IE &= ~BIT0;

//		 prova=(P2IES & 0x0001);

		 if (*FLAG_interrupt==3){

	         TA1CCTL0 = 0;  // Disable Interrupt
	    	 TA1CTL = 0;    // Disable Timer

			 Counter_3= Counter_3 + 1;

/*

			 StartDebounceTimer_2();
				// LPM4
			 __bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
			 __no_operation();

*/

//			 vmonitor_3=3400+(Counter_3*100);  //minimun value

			 unsigned int vmonitor_4;

			 vmonitor_4= VMIN + (Counter_3*100);  //minimun value

			 Set_V(vmonitor_4); // Vmonitor by using the external circuit

			 t=1;

			 t = (P2IN & 0x0001);

			 if(t==0){ //the voltage is still increasing slowly

				*FLAG_interrupt=3;

				P2IFG = 0;                         // P2 IFG flag cleared

				P2IES &= ~(BIT0);                  // P2.0 Lo/Hi edge interrupt P2OUT &= ~(BIT0);

				P2IE |= BIT0;                      // P2.0 interrupt enabled

				P2IFG = 0;                         // P2 IFG flag cleared

				// Port Configuration

				//This Pin P1.4 is not working properly. don't use.

				P1OUT &= ~(BIT4);
				P1DIR |= BIT4;



				StartDebounceTimer_2();

				// LPM4
				__bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
				__no_operation();


			   }else{  //the voltage is increasing quickly

				   unsigned int vmonitor_8;

			    	  vmonitor_8=VMAX;  //minimun value

			          // set up external Comparator = Vmonitor

			    	   Set_V(vmonitor_8); // Vmonitor by using the external circuit

			    	   __delay_cycles(500);

			    	   t=1;

			    	   t = (P2IN & 0x0001);

			    	   if(t==0){  //the voltage is increasing slowly

			    	  		*FLAG_interrupt=1;


			    	  		P2IFG = 0;                         // P2 IFG flag cleared
			    	  		P2IES &= ~(BIT0);                  // P2.0 Lo/Hi edge interrupt P2OUT &= ~(BIT0);
			    	  		P2IE |= BIT0;                      // P2.0 interrupt enabled
			    	  		P2IFG = 0;                         // P2 IFG flag cleared

			    	  		// Port Configuration

			    	  		//This Pin P1.4 is not working properly. don't use.

			    	  		P1OUT &= ~(BIT4);
			    	  		P1DIR |= BIT4;

			    	  		StartDebounceTimer_2();

			    	  		// LPM4
			    	  		__bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
			    	  		__no_operation();

			    	   }

			      }

		 }

		 if (*FLAG_interrupt==2){

//			 if(prova==0){

			 P2IFG = 0;
			 P2IE &= ~BIT0;

//     		__delay_cycles(10000);                   //only way this interrupt works correctly

			 CheckPoint();

			 if(pro==0){

				    unsigned int vmonitor_5;

//				    vmonitor_5=VMIN + (Counter_3*100);

				    vmonitor_5 = VMIN + 600 + (Counter_3*100);

				    if(vmonitor_5>VMAX){

				    	vmonitor_5=VMAX;

				    }


				    Set_V(vmonitor_5); // Vmonitor by using the external circuit

					t=1;

					t = (P2IN & 0x0001);

					if(t==0){

						*FLAG_interrupt=4;

						P2IFG = 0;                         // P2 IFG flag cleared

						P2IES &= ~(BIT0);                  // P2.0 Lo/Hi edge interrupt P2OUT &= ~(BIT0);

						P2IE |= BIT0;                      // P2.0 interrupt enabled

						P2IFG = 0;                         // P2 IFG flag cleared

						// Port Configuration

						//This Pin P1.4 is not working properly. don't use.


//		             	P3DIR |= BIT6;
//		            	P3OUT |= BIT6; // LED  ON

						P1OUT &= ~(BIT4);
						P1DIR |= BIT4;

//						StartDebounceTimer_2();

//						if (*FLAG_interrupt==4){
						// LPM4
						__bis_SR_register(LPM4_bits + GIE);        // LPM0, ADC10_ISR will force exit
						__no_operation();

//						 }


					   }


			 }

		 }


//		        *FLAG_interrupt==2;

		 if (*FLAG_interrupt==4){

		   unsigned int value_7 = VMINN + deltaV;

//		   unsigned long int value_4 = VMIN2; //for testing

		   Set_V(value_7);

		   *FLAG_interrupt=2;

		   P2IFG = 0;                         // P2 IFG flag cleared

		   P2IES |= BIT0;                     // P2.0 Hi/lo edge interrupt

		   P2IE |= BIT0;                      // P2.0 interrupt enabled
		   P2IFG = 0;                         // P2 IFG flag cleared


//		   StartDebounceTimer_2();

           P2DIR |= BIT6;
           P2OUT |= BIT6; // LED  ON

		   __bis_SR_register(GIE);        // LPM0, ADC10_ISR will force exit
		   __no_operation();

//		   break;

		 }

			pro=0;

			__bic_SR_register_on_exit(LPM4_bits); // Exit LPM4



         break;

         default:
         break;
      }

  }


/**********************************************************************//**
 * @brief Timer A1 ISR for debounce Timer
 *
 * @param  none
 *
 * @return none
 *************************************************************************/
#pragma vector = TIMER1_A0_VECTOR
__interrupt void Timer1_A0_ISR(void)
{
  TA1CCTL0 = 0;  // Disable Interrupt
  TA1CTL = 0;    // Disable Timer

  // stop also the interrupt related to the threshold

  P2IFG = 0;
  P2IE &= ~BIT0;

  __bic_SR_register_on_exit(LPM4_bits); // Clear CPUOFF bit from 0 (SR)

}








