/*
Hibernus++ for MSP430FR5739:

Hibernus++: Software-based approach to hibernate and restore adaptively the system's state
in response to a power failure. This software exploits an external comparator.

Citation: If you are using this code for your research please cite:

[1]	D. Balsamo, A. S. Weddell, G. V. Merrett, B. M. Al-Hashimi, D. Brunelli and L. Benini,
"Hibernus: Sustaining Computation During Intermittent Supply for Energy-Harvesting Systems,"
in IEEE Embedded Systems Letters, vol. 7, no. 1, pp. 15-18, March 2015.

[2]	D. Balsamo; A. Weddell; A. Das; A. Arreola; D. Brunelli; B. Al-Hashimi; G. Merrett; L. Benini,
"Hibernus++: A Self-Calibrating and Adaptive System for Transiently-Powered Embedded Devices,"
in IEEE Transactions on Computer-Aided Design of Integrated Circuits and Systems. 

This research has been supported by the University of Southampton and the University of Bologna.
Copyright 2018, Domenico Balsamo, All rights reserved.
06/02/2019
*/



#include <stdio.h>

// The FRAM section from 0xD000 - 0xF000 is used by all modes
// for performing writes to FRAM
// Do not use this section for code or data placement.
// It will get overwritten!

#define FLAG_3 0xEDC0
#define COUNT_3 0xEDC4
#define COUNT_2 0xEDC8
#define FLAG_2 0xEDCC
#define VMON_3 0xEDD0
#define VMON_2 0xEDD4
#define INT 0xEDD8
#define CHECK 0xEDDC
#define CALIBR 0xEDE0
#define COUNT 0xEDE4
#define FLAG 0xEDE8
#define VMON 0xEDEC


#define FRAM_START 0xEDF0
#define FRAM_END 0xFF40

// All RAM

#define RAM_START 0x1C00
#define RAM_END 0x2000 //All RAM

// Test

#define FRAM_TEST_START 0xD400
#define FRAM_TEST_END 0xE155
#define MEM_UNIT 0x200

//Voltage profile

//#define VMIN 2000
//#define VMINN 1880
//#define VMAX 3600

//Current profile

#define VMIN 2300
#define VMINN 2000
#define VMAX 3600

//debugger profile

//#define VMIN 3400
//#define VMIN2 3400
//#define VMAX 3600


// All Stack

#define PC 0xEDF0


#define V_THRESH 2200 // 2.2 volts


// Function Declarations
extern void SystemInit(void);
extern void StartUpSequence(void);
extern void LongDelay(void);
extern void CheckPoint (void);
extern void VoltageCheck (void);
extern void Restore(void);
extern void Save_RAM (void);
extern void Save_Register (void);
extern void Restore_RAM (void);
extern void Restore_Register (void);
extern void StartDebounceTimer(void);
extern void StartDebounceTimer_2(void);
extern void Restore_Comparator_State(void);
extern void Set_V(unsigned int);
extern void Calibration(void);
extern void Vmonitor(void);
extern void Vmonitor_2(void);
extern unsigned int Calibrate(void);
extern void ADCSetup(void);
extern void test(void);
extern void CheckPoint_calibrate (void);
extern void FRAM_Read(unsigned int StartAddress);
extern void GPIO_configuration(void);
extern void Clock_configuration(void);
extern void Test(void);


extern volatile unsigned long int ADCResult = 0;
